unit ldn189u1;

{$mode objfpc}{$H+}


//This demo application, LDN_189, demonstrates loading a
//  StringGrid from a CSV file. LDN_189 is explained in
//      http://sheepdogguides.com/lut/ltn4d.htm

//A much bigger application, which uses the code in
//  LDN_189, discusses using a StringGrid to access a
//  CSV file, make checks on the validity of data in that
//  file (for instance, one field is checked to see if
//  it contains a valid abbreviation for a US state. Any
//  invalid datum is highlighted, and can be edited.)
//That application is discussed in the tutorial at...
//      http://sheepdogguides.com/lut/ltn4c.htm

//N.B.: At the moment, there is a Kludge... search on
//   bColIndexL>8
//   ... to find it. It limits the number of fields which
//   can be in a record.

//Author's notes to himself: At bottom.

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Grids, ClipBrd;

const vers='1 Oct 16';
   //started 29 Sep 16, on road to an application to
   //   vett DVD records data.
   kGridMaxXIndex=15;
   kGridMaxYIndex=25;

type

  { Tldn189f1 }
  Tldn189f1 = class(TForm)
    buQuit: TButton;
    buLoadSGWay: TButton;
    buAdjColWidth: TButton;
    buAbout: TButton;
    buLoadLazWay: TButton;
    buLinkToWeb: TButton;
    sgData: TStringGrid;
    procedure buAboutClick(Sender: TObject);
    procedure buLinkToWebClick(Sender: TObject);
    procedure buLoadSGWayClick(Sender: TObject);
    procedure buQuitClick(Sender: TObject);
    procedure buAdjColWidthClick(Sender: TObject);
    procedure buLoadLazWayClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);

  private
    { private declarations }
    procedure LoadFileToStringGrid(sFileName:string);
    procedure StringToSGRow(sgLocal:TStringGrid;
          sRawData:string;iRow:integer;boTrimFields:boolean);
  public
    { public declarations }
  end;

var
  ldn189f1: Tldn189f1;

implementation

{$R *.lfm}
{$R+}

{ Tldn189f1 }

procedure Tldn189f1.buQuitClick(Sender: TObject);
begin
  close;
end;

procedure Tldn189f1.buAdjColWidthClick(Sender: TObject);
begin
  sgData.AutoSizeColumns;
end;

procedure Tldn189f1.buLoadLazWayClick(Sender: TObject);
begin
  //This may be "simple", but, sadly, it treats spaces as
  //field separators, unless the whole field is
  //enclosed in quotes. The last parameter has to do
  //with whether there's a line of headers in the CSV
  //
  //This "answer" sets the number of columns in the
  //StringGrid to the number of fields in the first
  //record (and throws away "extras" on subsequent
  //lines.) (My "answer" WILL throw away extras, too,
  //but, at 30 Sep 16, set the number of columns
  //crudely. (At present, "extras" cause problems.)
  //
  //My "answer" could be enhanced to set the number
  //of colums as LoadFromCSVFile does.
  //
  //Both answers treat...
  //  My,Data,
  //as a line with THREE fields, the last holding ''
  //I.e. the third field is "nothing" which is DIFFERENT
  //from "not specified".
  //
  //I'm not sure that this would work in Delphi, or, if it
  //does, will work exactly as it does in Lazarus.
  sgData.LoadFromCSVFile('TestDataLDN189.txt',',',true);
end;

procedure Tldn189f1.LoadFileToStringGrid(sFileName:string);
(*The file to be loaded should consist of one or more
  RECORDS, each consisting of one or more FIELDS.
  The file should consist printable characters, plus
  (maybe) the code for TAB, and a FEW others like that.
  The RECORDS are separated by CR/LF terminators,
  FIELDS are separated by COMMAS. A line ending in
  a comma is considered to indicate that the final
  field of that record had nothing in it, e.g.....
    Three fields: Fred,Bloggs,860-767-5555
    Four fields: Fred,Bloggs,860-767-5555,
  *)

var
  cardCounter:Cardinal;
  stringsTmp:TStrings;
begin //LoadFileToStringGrid
  stringsTmp:=TStringList.Create;
  try
    stringsTmp.LoadFromFile(sFileName);
    sgData.RowCount:=stringsTmp.Count;//Yes: .count, not .count-1
    sgData.ColCount:=7;//This CLUMSY... But could fairly easily
      //be replaced by code to count the fields in the first
      //line, and make the StringGrid have that many columns.
      //The "built in" LoadFromCSVFile does that. Both routines
      //ignore the fact that a data file might have more
      //fields in a subsequent record. (The extra fields are
      //just thrown away in the LoadFromCSVFile. At present,
      //they cause a problem (good! You'll know about them!)
      //in this answer. (That can be fixed.)
      //What's done here, for the moment, is to set, by hand, something
      //that limits how many fields there can be in any record
      //in the CSV file. Exceed that limit, and the application
      //shuts down, in a messy way. Fall short of that on a given
      //line, and you just have some empty cells at the right hand
      //end of the line.

      //For the future: Add a parameter to LoadFileToStringGrid
      //If zero: Use the "answer" used by LoadFromCSVFile, if
      //> zero, set ColCount from that. (If this is done, also
      //provide a "scan file, find out longest (most fields) record"
      //function.)
      //
      //In any case, add a "var" variable to return error codes
      //to user.... "file not found","too many fields encountered",
      //etc.

      //Also add a parameter to specify the separator character.
      //Add a way to use....  "Bloggs, Fred",123 to allow separator
      //  character inside a field? Make it optional, and tell
      //users that the application runs faster if they don't
      //use the feature?

    for cardCounter:= 0 to stringsTmp.Count-1 do begin
          StringToSGRow(sgData,stringsTmp[cardCounter],cardCounter,true);
          end;

      //To be checked... what does app do when a)there is / b) there
      //  is not a CR at the end of the file? (Prelim checks: It doesn't
      //  matter.  (^_^)
      //If there is a last line consisting just of spaces, the Lazarus
      //  version does NOT add a row to the StringGrid, the
      //  SheepdogGuides version DOES add a row

  finally
    stringsTmp.Free;
    end;//Try... finally

  //Put....  sgData.AutoSizeColumns;
  //   ... in your code, after the call of LoadFileToStringGrid,
  //   if you would like all of the columns' widths adjusted,
  //   leaving each column just wide enough for the widest datum
  //   in the column.

end;//LoadFileToStringGrid

procedure Tldn189f1.StringToSGRow(sgLocal:TStringGrid;
                             sRawData:string;iRow:integer;boTrimFields:boolean);
var bColIndexL:byte;
  sOneField:string;
  boDoneIt:boolean;
//This procedure is used within "LoadFileToStringGrid". You
//  wouldn't be likely to want to call it directly.

//Takes, from sRawData, a string like...
//  860-555-1912,CT,Joe Smith
//... and puts the three fields in the first three
//colums of StringGrid sgLocal.
//What's in sRawData is "eaten away" over the course of the procedure.
//If boTrimFields is true, then spaces at either end of the field datum
//  are trimmed off....
//  "   sample datum   "
//  ... would become...
//  "sample datum"

  function boChomp(var sSource:string;
        var sDest:string):boolean;//SR of StringToSGRow
  //N.B. BOTH PARAMS ARE *var* PARAMS... contents of the
  //variables used to "feed" this SR will be changed by
  //the execution of the procedure.
  //
  //The contents of sDest before the call are irrelevant.
  //
  //If the string passed to sSource has a comma at it's
  //  right hand end, then the routine assumes that to mean
  //  that there is a field after the comma, but that the
  //  contents of that field just happen to be "nothing",
  //  i.e. ''. In the processing of such a string, sSource
  //  will be passed back to the calling program with a
  //  rogue value, just before the call of Chomp which
  //  "reads" the "empty" field.
  //In other words... string "a" has three fields,
  //  string "b" has FOUR...
  //     a) Fred,Bloggs,860-767-5555
  //     b) Fred,Bloggs,860-767-5555,
  //                             ---^---
  var iPosOfComma:integer;
  const kFieldSep=',';//Rewrite to supply this in
     //a parameter, if you feel having that
     //flexiblity is important to you.

  begin //main block of boChomp, SR of StringToSGRow
    if sSource='' then result:=true//no ; here
         //(See "N.B. bChomp returns...", below. THIS "true"
         //arises if you call Chomp when there was no point...
         //you already knew there was no more data to parse.
    else begin //1  (Use last field in string that started
         //String1, String2, .... , StringLast
      //(There will be no comma after last field.)

      //N.B. boChomp returns false until there is no more
      //  data to be harvested by a further call of Chomp.
      //Note especially... the first time it
      //  returns TRUE, there is STILL one field to be
      //  USED by the calling program. It has been
      //  returned to the calling program in sDest,
      //  as usual.

      if sSource='tkbRogueToIndicateNullFieldAtEndOfRaw' then
            sSource:='';
      iPosOfComma:=pos(kFieldSep,sSource);
      if iPosOfComma=0 then begin //2
           //The "=1" case arises if the string you are
           //chomping ends with a comma, which is interpreted
           //as meaning that there IS one more field in the
           //string, but it just happens to consist of nothing,
           //just happens to be ''}
        sDest:=sSource;
        sSource:='';
        result:=true;//(See "N.B. bChomp returns..." THIS "true"
          //is for the case where you have just harvested the last
          //field from the string you have been "chomping" through.
        end//no ; here. End of "then 2"
      else begin //2
          if iPosOfComma=length(sSource) then
               sSource:=sSource+'tkbRogueToIndicateNullFieldAtEndOfRaw';
          sDest:=copy(sSource,1,iPosOfComma-1);
          sSource:=copy(sSource,iPosOfComma+1,length(sSource));
          result:=false;
          end;//of "else 2"
        end;// of "else 1"
  end;//boChomp, SR of StringToSGRow

begin //main block of StringToSGRow
   bColIndexL:=0;
   boDoneIt:=false;
   repeat
     boDoneIt:=boChomp(sRawData,sOneField);//N.B: These are "var"
          //parameters. A bit is chopped off of sRawData, and
          //sOneField is filled with a new value. Also: You can
          //ignore compiler warning about sOneField not being
          //initialized.
     if boTrimfields then sOneField:=trim(sOneField);

     //If you want to "see something" when a field in the
     //  CSV says "nothing, here", de-rem the following...
     //if sOneField='' then sOneField:='null';
     sgLocal.cells[bColIndexL,iRow]:=sOneField;

     inc(bColIndexL);

   until (boDoneIt) OR (bColIndexL>8);//second term to be refined in
     //due course... at present, it is a kludge... but it WILL have
     //a role to play one day. Keep note at top of code in step.
     //(Eventually, before LoadFileToStringGrid,
     //is called, somehow a determination will be made as to how many
     //columns are to be filled. If, by chance or error, the CSV file
     //has a record with too many fields, at the moment problems arise.
     //When the "8" is replaced by a variable, and that is loaded to
     //reflect how big the StringGrid available is, then steps can be
     //taken to deal nicely with a "too big for the row" record from
     //the CSV file.
end;//StringToSGRow

procedure Tldn189f1.FormCreate(Sender: TObject);
begin
  ldn189f1.caption:='LDN189- Load CSV into StringGrid- ver: '+vers;
  application.title:='LDN189';
end;

procedure Tldn189f1.buLoadSGWayClick(Sender: TObject);
begin
LoadFileToStringGrid('TestDataLDN189.txt');
end;

procedure Tldn189f1.buAboutClick(Sender: TObject);
begin
  showmessage('This is one of many applications '+
    'available from http://SheepdogGuides.com/Lut/'+chr(13)+
    '... a collection of Lazarsus and Delphi tutorials.');
end;

procedure Tldn189f1.buLinkToWebClick(Sender: TObject);
begin
  Clipboard.AsText:='http://sheepdogguides.com/lut/lt4Nd.htm';
end;

(*Author's notes to himself...

LDN_189 derived from LDN_190, itself derived from LDN191!

LDN_191 is a complex application illustrating Doing Clever
  Things with a StringGrid object. Sadly, at it's heart,
  30 Sep 16, the code for the basic loading of the
  StringGrid from data in a CSV file was flawed. I am
  01 Oct 16, working on fixing that.

LDN_189 is a demo for the new, working (I hope!) core
  "load CSV to StringGrid", believed to be fairly sound.
  It demos two ways to do the job.

Where LDN_191 says/ said...
       sgData.Rows[I-1].CommaText := CSV[I-1];
 .... needs (needed? Maybe the changes have been made,
         since this was written?)... needs to be changed to...
       StringToSGRow(sgData,CSV[I],I,true);
... and the code to support that needs to be added.
*)

end.

